# Copyright 2020-2021 Mufeed Ali
# Copyright 2020-2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import re
import threading
from gettext import gettext as _

from gi.repository import Gio, GLib, GObject, Gtk, Handy

from dialect.define import RES_PATH
from dialect.settings import Settings
from dialect.translators import TRANSLATORS
from dialect.tts import TTS


@Gtk.Template(resource_path=f'{RES_PATH}/preferences.ui')
class DialectPreferencesWindow(Handy.PreferencesWindow):
    __gtype_name__ = 'DialectPreferencesWindow'

    parent = NotImplemented

    # Get preferences widgets
    dark_mode = Gtk.Template.Child()
    live_translation = Gtk.Template.Child()
    translate_accel = Gtk.Template.Child()
    src_auto = Gtk.Template.Child()
    backend = Gtk.Template.Child()
    backend_instance = Gtk.Template.Child()
    backend_instance_row = Gtk.Template.Child()
    backend_instance_stack = Gtk.Template.Child()
    backend_instance_label = Gtk.Template.Child()
    backend_instance_edit = Gtk.Template.Child()
    backend_instance_save = Gtk.Template.Child()
    backend_instance_reset = Gtk.Template.Child()
    backend_instance_edit_box = Gtk.Template.Child()
    tts = Gtk.Template.Child()
    tts_row = Gtk.Template.Child()
    search_provider = Gtk.Template.Child()

    def __init__(self, parent, **kwargs):
        super().__init__(**kwargs)

        self.parent = parent

        self.setup()

    def setup(self):
        # Disable search, we have few preferences
        self.set_search_enabled(False)
        # Temporal fix for crash
        self.connect('destroy', self._unbind_settings)

        # Setup translate accel combo row
        model = Gio.ListStore.new(Handy.ValueObject)
        options = ['Ctrl + Enter', 'Enter']
        for index, value in enumerate(options):
            model.insert(index, Handy.ValueObject.new(value))
        self.translate_accel.bind_name_model(model,
                                             Handy.ValueObject.dup_string)

        # Setup backends combo row
        self.backend_model = Gio.ListStore.new(BackendObject)
        backend_options = [
            BackendObject(translator.name, translator.prettyname) for translator in TRANSLATORS.values()
        ]
        selected_backend_index = 0
        for index, value in enumerate(backend_options):
            self.backend_model.insert(index, value)
            if value.name == Settings.get().backend:
                selected_backend_index = index
        self.backend.bind_name_model(self.backend_model,
                                     BackendObject.get_name)

        # Bind preferences with GSettings
        Settings.get().bind('dark-mode', self.dark_mode, 'active',
                            Gio.SettingsBindFlags.DEFAULT)
        Settings.get().bind('live-translation', self.live_translation, 'active',
                            Gio.SettingsBindFlags.DEFAULT)
        Settings.get().bind('translate-accel', self.translate_accel,
                            'selected-index', Gio.SettingsBindFlags.DEFAULT)
        Settings.get().bind('src-auto', self.src_auto, 'active',
                            Gio.SettingsBindFlags.DEFAULT)

        # Setup TTS
        self.tts_row.set_visible(len(TTS) >= 1)
        self.tts.set_active(Settings.get().tts != '')

        # Toggle dark mode
        self.dark_mode.connect('notify::active', self._toggle_dark_mode)

        # Set translate accel sensitivity by live translation state
        self.translate_accel.set_sensitive(not self.live_translation.get_active())
        self.live_translation.connect('notify::active', self._toggle_accel_pref)

        # Switch backends
        self.backend.set_selected_index(selected_backend_index)
        self.backend.connect('notify::selected-index', self._switch_backends)
        self.parent.connect('notify::backend-loading', self._on_backend_loading)

        # Toggle TTS
        self.tts.connect('notify::active', self._toggle_tts)

        # Change translator instance
        Settings.get().connect('changed', self._on_settings_changed)
        self.backend_instance_edit.connect('clicked', self._on_edit_backend_instance)
        self.backend_instance_save.connect('clicked', self._on_save_backend_instance)
        self.backend_instance_reset.connect('clicked', self._on_reset_backend_instance)
        self.__check_instance_support()

        self.instance_save_image = Gtk.Image.new_from_icon_name(
            'emblem-ok-symbolic', Gtk.IconSize.BUTTON)
        self.backend_instance_save.add(self.instance_save_image)
        self.instance_save_spinner = Gtk.Spinner()
        self.instance_save_image.show()
        self.instance_save_spinner.show()

        self.error_popover = Gtk.Popover(
            relative_to=self.backend_instance, can_focus=False, modal=False)
        self.error_label = Gtk.Label(label='Not a valid instance')
        error_icon = Gtk.Image.new_from_icon_name(
            'dialog-error-symbolic', Gtk.IconSize.LARGE_TOOLBAR)
        error_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, margin=8, spacing=8)
        error_box.pack_start(error_icon, False, False, 0)
        error_box.pack_start(self.error_label, False, False, 0)
        self.error_popover.add(error_box)
        self.error_popover.set_position(Gtk.PositionType.BOTTOM)
        error_box.show_all()
        self.error_popover.hide()

        # Search Provider
        if os.getenv('XDG_CURRENT_DESKTOP') != 'GNOME':
            self.search_provider.hide()

    def _unbind_settings(self,  *args, **kwargs):
        Settings.get().unbind(self.dark_mode, 'active')
        Settings.get().unbind(self.live_translation, 'active')
        Settings.get().unbind(self.src_auto, 'active')

    def _on_settings_changed(self, _settings, key):
        backend = Settings.get().backend
        if key == 'backend-settings':
            if TRANSLATORS[backend].supported_features['change-instance']:
                # Update backend
                Settings.get().reset_src_langs(backend)
                Settings.get().reset_dest_langs(backend)
                self.parent.change_backends(backend)

    def _toggle_dark_mode(self, switch, _active):
        gtk_settings = Gtk.Settings.get_default()
        active = switch.get_active()
        gtk_settings.set_property('gtk-application-prefer-dark-theme', active)

    def _toggle_accel_pref(self, switch, _active):
        self.translate_accel.set_sensitive(not switch.get_active())

    def _toggle_tts(self, switch, _active):
        value = ''
        if switch.get_active() and len(TTS) >= 1:
            tts = list(TTS.keys())
            value = str(tts[0])

        self.parent.src_voice_btn.set_sensitive(False)
        self.parent.src_voice_btn.set_visible(switch.get_active())
        self.parent.dest_voice_btn.set_sensitive(False)
        self.parent.dest_voice_btn.set_visible(switch.get_active())

        Settings.get().tts = value

        if switch.get_active():
            threading.Thread(
                target=self.parent.load_lang_speech,
                daemon=True
            ).start()

    def _switch_backends(self, row, _value):
        backend = self.backend_model[row.get_selected_index()].name
        Settings.get().backend = backend
        self.__check_instance_support()
        self.parent.change_backends(backend)

    def _on_backend_loading(self, window, _value):
        self.backend.set_sensitive(not window.get_property('backend-loading'))
        self.backend_instance_row.set_sensitive(not window.get_property('backend-loading'))

    def _on_edit_backend_instance(self, _button):
        backend = Settings.get().backend
        self.backend_instance_stack.set_visible_child_name('edit')
        self.backend_instance.set_text(Settings.get().get_instance_url(backend))

    def _on_save_backend_instance(self, _button):
        backend = Settings.get().backend
        old_value = Settings.get().get_instance_url(backend)
        new_value = self.backend_instance.get_text()

        url = re.compile(r"https?://(www\.)?")
        new_value = url.sub('', new_value).strip().strip('/')

        if new_value != old_value:
            # Validate
            threading.Thread(
                target=self.__validate_new_backend_instance,
                args=[new_value],
                daemon=True
            ).start()
        else:
            self.backend_instance_stack.set_visible_child_name('view')

    def _on_reset_backend_instance(self, _button):
        backend = Settings.get().backend
        Settings.get().reset_instance_url(backend)
        self.backend_instance_stack.set_visible_child_name('view')
        Gtk.StyleContext.remove_class(self.backend_instance.get_style_context(), 'error')
        self.error_popover.popdown()

    def __check_instance_support(self):
        backend = Settings.get().backend
        if TRANSLATORS[backend].supported_features['change-instance']:
            self.backend_instance_row.set_visible(True)
            self.backend_instance_label.set_label(Settings.get().get_instance_url(backend))
        else:
            self.backend_instance_row.set_visible(False)

    def __validate_new_backend_instance(self, url):
        def spinner_start():
            self.backend.set_sensitive(False)
            self.backend_instance_row.set_sensitive(False)
            self.backend_instance_save.remove(self.instance_save_image)
            self.backend_instance_save.add(self.instance_save_spinner)
            self.instance_save_spinner.start()

        def spinner_end():
            self.backend.set_sensitive(True)
            self.backend_instance_row.set_sensitive(True)
            self.backend_instance_save.remove(self.instance_save_spinner)
            self.backend_instance_save.add(self.instance_save_image)
            self.backend_instance_label.set_label(Settings.get().get_instance_url(backend))
            self.instance_save_spinner.stop()

        GLib.idle_add(spinner_start)
        backend = Settings.get().backend
        validate = TRANSLATORS[backend].validate_instance_url(url)
        if validate:
            Settings.get().set_instance_url(backend, url)
            GLib.idle_add(Gtk.StyleContext.remove_class, self.backend_instance.get_style_context(), 'error')
            GLib.idle_add(self.backend_instance_stack.set_visible_child_name, 'view')
            GLib.idle_add(self.error_popover.popdown)
        else:
            GLib.idle_add(Gtk.StyleContext.add_class, self.backend_instance.get_style_context(), 'error')
            error_text = _('Not a valid {backend} instance')
            error_text = error_text.format(backend=TRANSLATORS[backend].prettyname)
            GLib.idle_add(self.error_label.set_label, error_text)
            GLib.idle_add(self.error_popover.popup)

        GLib.idle_add(spinner_end)


class BackendObject(GObject.Object):
    name = None
    prettyname = None

    def __init__(self, name, prettyname):
        super().__init__()
        self.name = name
        self.prettyname = prettyname

    def get_name(self):
        return self.prettyname



# Copyright 2020 gi-lom
# Copyright 2020-2021 Mufeed Ali
# Copyright 2020-2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

# Initial setup
import sys
from gettext import gettext as _

import gi
gi.require_version('Gdk', '3.0')
gi.require_version('Gtk', '3.0')
gi.require_version('Gst', '1.0')
gi.require_version('Handy', '1')

from gi.repository import Gdk, Gio, GLib, Gst, Gtk, Handy

from dialect.define import APP_ID, RES_PATH
from dialect.preferences import DialectPreferencesWindow
from dialect.settings import Settings
from dialect.window import DialectWindow


class Dialect(Gtk.Application):
    def __init__(self, version):
        Gtk.Application.__init__(
            self,
            application_id=APP_ID,
            flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE
        )

        # App window
        self.version = version
        self.window = None
        self.launch_text = ''
        self.launch_langs = {}

        # Add command line options
        self.add_main_option('text', b't', GLib.OptionFlags.NONE,
                             GLib.OptionArg.STRING, 'Text to translate', None)
        self.add_main_option('src', b's', GLib.OptionFlags.NONE,
                             GLib.OptionArg.STRING, 'Source lang code', None)
        self.add_main_option('dest', b'd', GLib.OptionFlags.NONE,
                             GLib.OptionArg.STRING, 'Destination lang code', None)

        self.setup_actions()

    def do_activate(self):
        self.window = self.props.active_window

        if not self.window:
            width, height = Settings.get().window_size
            self.window = DialectWindow(
                application=self,
                # Translators: Do not translate the app name!
                title=_('Dialect'),
                default_height=height,
                default_width=width,
                text=self.launch_text,
                langs=self.launch_langs
            )
            self.setup_actions_signals()

        self.window.present()

    def do_command_line(self, command_line):
        options = command_line.get_options_dict()
        options = options.end().unpack()
        text = ''
        langs = {
            'src': None,
            'dest': None
        }

        if 'text' in options:
            text = options['text']
        if 'src' in options:
            langs['src'] = options['src']
        if 'dest' in options:
            langs['dest'] = options['dest']

        if self.window is not None:
            self.window.translate(text, langs['src'], langs['dest'])
        else:
            self.launch_text = text
            self.launch_langs = langs

        self.activate()
        return 0

    def do_startup(self):
        Gtk.Application.do_startup(self)
        GLib.set_application_name(_('Dialect'))
        GLib.set_prgname('com.github.gi_lom.dialect')

        Handy.init()  # Init Handy
        Gst.init(None)  # Init Gst

        # Load CSS
        css_provider = Gtk.CssProvider()
        css_provider.load_from_resource(f'{RES_PATH}/style.css')
        screen = Gdk.Screen.get_default()
        style_context = Gtk.StyleContext()
        style_context.add_provider_for_screen(screen, css_provider, Gtk.STYLE_PROVIDER_PRIORITY_USER)

    def setup_actions(self):
        """ Setup menu actions """
        self.pronunciation_action = Gio.SimpleAction.new_stateful(
            'pronunciation', None, Settings.get().show_pronunciation_value
        )
        self.set_accels_for_action('app.pronunciation', ['<Primary>P'])
        self.add_action(self.pronunciation_action)

        self.preferences_action = Gio.SimpleAction.new('preferences', None)
        self.set_accels_for_action('app.preferences', ['<Primary>comma'])
        self.add_action(self.preferences_action)

        self.shortcuts_action = Gio.SimpleAction.new('shortcuts', None)
        self.set_accels_for_action('app.shortcuts', ['<Primary>question'])
        self.add_action(self.shortcuts_action)

        self.about_action = Gio.SimpleAction.new('about', None)
        self.add_action(self.about_action)

        self.back_action = Gio.SimpleAction.new('back', None)
        self.back_action.set_enabled(False)
        self.set_accels_for_action('app.back', ['<Alt>Left'])
        self.add_action(self.back_action)

        self.forward_action = Gio.SimpleAction.new('forward', None)
        self.forward_action.set_enabled(False)
        self.set_accels_for_action('app.forward', ['<Alt>Right'])
        self.add_action(self.forward_action)

        self.switch_action = Gio.SimpleAction.new('switch', None)
        self.set_accels_for_action('app.switch', ['<Primary>S'])
        self.add_action(self.switch_action)

        self.clear_action = Gio.SimpleAction.new('clear', None)
        self.clear_action.set_enabled(False)
        self.set_accels_for_action('app.clear', ['<Primary>D'])
        self.add_action(self.clear_action)

        self.paste_action = Gio.SimpleAction.new('paste', None)
        self.set_accels_for_action('app.paste', ['<Primary><Shift>V'])
        self.add_action(self.paste_action)

        self.copy_action = Gio.SimpleAction.new('copy', None)
        self.copy_action.set_enabled(False)
        self.set_accels_for_action('app.copy', ['<Primary><Shift>C'])
        self.add_action(self.copy_action)

        self.listen_dest_action = Gio.SimpleAction.new('listen-dest', None)
        self.set_accels_for_action('app.listen-dest', ['<Primary>L'])
        self.add_action(self.listen_dest_action)

        self.listen_src_action = Gio.SimpleAction.new('listen-src', None)
        self.set_accels_for_action('app.listen-src', ['<Primary><Shift>L'])
        self.add_action(self.listen_src_action)

        self.quit_action = Gio.SimpleAction.new('quit', None)
        self.set_accels_for_action('app.quit', ['<Primary>Q'])
        self.add_action(self.quit_action)

    def setup_actions_signals(self):
        self.pronunciation_action.connect('change-state', self.on_pronunciation)
        self.preferences_action.connect('activate', self.on_preferences)
        self.shortcuts_action.connect('activate', self.on_shortcuts)
        self.about_action.connect('activate', self.on_about)
        self.back_action.connect('activate', self.window.ui_return)
        self.forward_action.connect('activate', self.window.ui_forward)
        self.switch_action.connect('activate', self.window.ui_switch)
        self.clear_action.connect('activate', self.window.ui_clear)
        self.paste_action.connect('activate', self.window.ui_paste)
        self.copy_action.connect('activate', self.window.ui_copy)
        self.listen_dest_action.connect('activate', self.window.ui_dest_voice)
        self.listen_src_action.connect('activate', self.window.ui_src_voice)
        self.quit_action.connect('activate', self.on_quit)

    def on_pronunciation(self, action, value):
        """ Update show pronunciation setting """
        action.set_state(value)
        Settings.get().show_pronunciation = value

        # Update UI
        if self.window.trans_src_pron is not None:
            self.window.src_pron_revealer.set_reveal_child(value)
        if self.window.trans_dest_pron is not None:
            self.window.dest_pron_revealer.set_reveal_child(value)

    def on_preferences(self, _action, _param):
        """ Show preferences window """
        window = DialectPreferencesWindow(self.window)
        window.set_transient_for(self.window)
        window.present()

    def on_shortcuts(self, _action, _param):
        """Launch the Keyboard Shortcuts window."""
        builder = Gtk.Builder.new_from_resource(f'{RES_PATH}/shortcuts-window.ui')
        translate_shortcut = builder.get_object('translate_shortcut')
        translate_shortcut.set_visible(not Settings.get().live_translation)
        translate_shortcut.set_property('accelerator', Settings.get().translate_accel)
        shortcuts_window = builder.get_object('shortcuts')
        shortcuts_window.set_transient_for(self.window)
        shortcuts_window.show()

    def on_about(self, _action, _param):
        """ Show about dialog """
        builder = Gtk.Builder.new_from_resource(f'{RES_PATH}/about.ui')
        about = builder.get_object('about')
        about.set_transient_for(self.window)
        about.set_logo_icon_name(APP_ID)
        about.set_version(self.version)
        about.connect('response', lambda dialog, response: dialog.destroy())
        about.present()

    def on_quit(self, _action, _param):
        self.quit()


def main(version):
    # Run the Application
    app = Dialect(version)
    return app.run(sys.argv)



# Copyright 2021 Mufeed Ali
# Copyright 2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

class TranslatorBase:
    name = ''
    prettyname = ''
    history = []
    languages = ['en']
    supported_features = {
        'mistakes': False,
        'pronunciation': False,
        'change-instance': False,
    }
    instance_url = ''
    src_langs = ['en', 'fr', 'es', 'de']
    dest_langs = ['fr', 'es', 'de', 'en']

    @staticmethod
    def validate_instance_url(url):
        pass

    def detect(self, src_text):
        pass

    def translate(self, src_text, src, dest):
        pass


class TranslationError(Exception):
    """Exception raised when translation fails."""

    def __init__(self, cause, message='Translation has failed'):
        self.cause = cause
        self.message = message
        super().__init__(self.message)


class Translation:
    text = None
    extra_data = {
        'possible-mistakes': None,
        'src-pronunciation': None,
        'dest-pronunciation': None,
    }

    def __init__(self, text, extra_data):
        self.text = text
        self.extra_data = extra_data


class Detected:
    lang = ''
    confidence = 0.0

    def __init__(self, lang, confidence):
        self.lang = lang
        self.confidence = confidence



# Copyright 2021 Mufeed Ali
# Copyright 2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

from googletrans import Translator as GoogleTranslator

from dialect.translators.basetrans import TranslatorBase, TranslationError, Translation


class Translator(TranslatorBase):
    name = 'google'
    prettyname = 'Google Translate'
    history = []
    languages = [
        'af', 'sq', 'am', 'ar', 'hy', 'az', 'eu', 'be', 'bn', 'bs', 'bg', 'ca',
        'ceb', 'ny', 'zh-CN', 'zh-TW', 'co', 'hr', 'cs', 'da', 'nl', 'en', 'eo',
        'et', 'tl', 'fi', 'fr', 'fy', 'gl', 'ka', 'de', 'el', 'gu', 'ht', 'ha',
        'haw', 'iw', 'he', 'hi', 'hmn', 'hu', 'is', 'ig', 'id', 'ga', 'it',
        'ja', 'jw', 'kn', 'kk', 'km', 'ko', 'ku', 'ky', 'lo', 'la', 'lv', 'lt',
        'lb', 'mk', 'mg', 'ms', 'ml', 'mt', 'mi', 'mr', 'mn', 'my', 'ne', 'no',
        'or', 'ps', 'fa', 'pl', 'pt', 'pa', 'ro', 'ru', 'sm', 'gd', 'sr', 'st',
        'sn', 'sd', 'si', 'sk', 'sl', 'so', 'es', 'su', 'sw', 'sv', 'tg', 'ta',
        'te', 'th', 'tr', 'uk', 'ur', 'ug', 'uz', 'vi', 'cy', 'xh', 'yi', 'yo',
        'zu'
    ]
    supported_features = {
        'mistakes': True,
        'pronunciation': True,
        'change-instance': False,
    }

    def __init__(self, **kwargs):
        self._translator = GoogleTranslator(raise_exception=True)

    def detect(self, src_text):
        try:
            if callable(getattr(self._translator, 'detect_legacy', None)):
                return self._translator.detect_legacy(src_text)
            return self._translator.detect(src_text)
        except Exception as exc:
            raise TranslationError(exc) from exc

    def translate(self, src_text, src, dest):
        try:
            if callable(getattr(self._translator, 'translate_legacy', None)):
                translated = self._translator.translate_legacy(src_text, src=src, dest=dest)
            else:
                translated = self._translator.translate(src_text, src=src, dest=dest)
            result = Translation(
                translated.text,
                {
                    'possible-mistakes': translated.extra_data['possible-mistakes'],
                    'src-pronunciation': None,
                    'dest-pronunciation': None,
                },
            )
            try:
                result.extra_data['src-pronunciation'] = translated.extra_data['translation'][1][3]
            except IndexError:
                result.extra_data['src-pronunciation'] = None
            try:
                result.extra_data['dest-pronunciation'] = translated.extra_data['translation'][1][2]
            except IndexError:
                result.extra_data['dest-pronunciation'] = None
            return result
        except Exception as exc:
            raise TranslationError(exc) from exc



# Copyright 2020-2021 Mufeed Ali
# Copyright 2020-2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import re

from gi.repository import GObject, Gtk

from dialect.define import RES_PATH
from dialect.translators import get_lang_name


@Gtk.Template(resource_path=f'{RES_PATH}/lang-selector.ui')
class DialectLangSelector(Gtk.Popover):
    __gtype_name__ = 'DialectLangSelector'

    # Get widgets
    search = Gtk.Template.Child()
    scroll = Gtk.Template.Child()
    revealer = Gtk.Template.Child()
    recent_list = Gtk.Template.Child()
    separator = Gtk.Template.Child()
    lang_list = Gtk.Template.Child()

    # Propeties
    selected = GObject.Property(type=str)  # Key of the selected lang

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Connect popover closed signal
        self.connect('closed', self._closed)
        # Connect list signals
        self.recent_list.connect('row-activated', self._activated)
        self.lang_list.connect('row-activated', self._activated)
        # Set filter func to lang list
        self.lang_list.set_filter_func(self.filter_func, None, False)
        # Connect search entry changed signal
        self.search.connect('changed', self._update_search)

    def filter_func(self, row, _data, _notify_destroy):
        search = self.search.get_text()
        return bool(re.search(search, row.name, re.IGNORECASE))

    def set_languages(self, languages):
        # Clear list
        children = self.lang_list.get_children()
        for child in children:
            self.lang_list.remove(child)

        # Load langs list
        for code in languages:
            row_selected = (code == self.selected)
            self.lang_list.insert(LangRow(code, get_lang_name(code), row_selected), -1)

    def insert_recent(self, code, name, position=-1):
        row_selected = (code == self.selected)
        self.recent_list.insert(LangRow(code, name, row_selected), position)

    def clear_recent(self):
        children = self.recent_list.get_children()
        for child in children:
            self.recent_list.remove(child)

    def refresh_selected(self):
        for lang in self.lang_list.get_children():
            lang.selected = (lang.code == self.selected)

    def _activated(self, _list, row):
        # Close popover
        self.popdown()
        # Set selected property
        self.set_property('selected', row.code)

    def _closed(self, _popover):
        # Reset scroll
        vscroll = self.scroll.get_vadjustment()
        vscroll.set_value(0)
        # Clear search
        self.search.set_text('')

    def _update_search(self, _entry):
        search = self.search.get_text()
        if search != '':
            self.revealer.set_reveal_child(False)
        else:
            self.revealer.set_reveal_child(True)
        self.lang_list.invalidate_filter()


class LangRow(Gtk.ListBoxRow):

    def __init__(self, code, name, selected=False, **kwargs):
        super().__init__(**kwargs)

        self.code = code
        self.name = name

        row_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=12)
        label = Gtk.Label(self.name,
                          halign=Gtk.Align.START,
                          margin_start=4)
        self.get_style_context().add_class('langselector')
        row_box.pack_start(label, False, True, 0)
        self.selected_icon = Gtk.Image.new_from_icon_name('object-select-symbolic', Gtk.IconSize.BUTTON)
        row_box.pack_start(self.selected_icon, False, True, 0)
        self.add(row_box)
        self.show_all()
        self.selected = selected

    @property
    def selected(self):
        return self.selected_icon.get_visible()

    @selected.setter
    def selected(self, value):
        self.selected_icon.set_visible(value)



# Copyright 2021 Mufeed Ali
# Copyright 2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import json

from gi.repository import Gio, GLib

from dialect.define import APP_ID
from dialect.translators import (
    check_backend_availability,
    get_fallback_backend_name,
    TRANSLATORS
)
from dialect.tts import TTS


class Settings(Gio.Settings):
    """
    Dialect settings handler
    """

    instance = None

    def __init__(self):
        Gio.Settings.__init__(self)

    @staticmethod
    def new():
        """Create a new instance of Settings."""
        g_settings = Gio.Settings.new(APP_ID)
        g_settings.__class__ = Settings
        return g_settings

    @staticmethod
    def get():
        """Return an active instance of Settings."""
        if Settings.instance is None:
            Settings.instance = Settings.new()

        return Settings.instance

    @property
    def window_size(self):
        value = self.get_value('window-size')
        return (value[0], value[1])

    @window_size.setter
    def window_size(self, size):
        width, height = size
        self.set_value('window-size', GLib.Variant('ai', [width, height]))

    @property
    def translate_accel(self):
        """Return the user's preferred translation shortcut."""
        value = self.translate_accel_value

        if value == 0:
            return '<Primary>Return'
        if value == 1:
            return 'Return'

        return '<Primary>Return'

    @property
    def translate_accel_value(self):
        """Return the user's preferred translation shortcut value."""
        return self.get_int('translate-accel')

    @property
    def tts(self):
        """Return the user's preferred TTS service."""
        value = self.get_string('tts-name')
        if value not in TTS.keys():
            value = ''
            self.tts = value
        return value

    @tts.setter
    def tts(self, value):
        """Set the user's preferred TTS service."""
        self.set_string('tts-name', value)

    @property
    def dark_mode(self):
        return self.get_boolean('dark-mode')

    @dark_mode.setter
    def dark_mode(self, state):
        self.set_boolean('dark-mode', state)

    @property
    def live_translation(self):
        return self.get_boolean('live-translation')

    @live_translation.setter
    def live_translation(self, state):
        self.set_boolean('live-translation', state)

    @property
    def show_pronunciation(self):
        return self.get_boolean('show-pronunciation')

    @property
    def show_pronunciation_value(self):
        return self.get_value('show-pronunciation')

    @show_pronunciation.setter
    def show_pronunciation(self, state):
        self.set_boolean('show-pronunciation', state)

    @property
    def src_auto(self):
        return self.get_boolean('src-auto')

    @src_auto.setter
    def src_auto(self, state):
        self.set_boolean('src-auto', state)

    @property
    def backend(self):
        """Return the user's preferred backend."""
        # Dialect 1.2.0 and below used the backend key and
        # stored the index of the chosen backend as an int.
        value = self.get_int('backend')

        if value == -1:
            value = self.get_string('backend-name')
        else:
            if value == 0:
                value = 'google'
            elif value == 1:
                value = 'libretranslate'

        if check_backend_availability(value):
            return value

        self.backend = get_fallback_backend_name()
        return get_fallback_backend_name()

    @backend.setter
    def backend(self, name):
        """
        Set the user's preferred backend.

        :param name: name of backend
        :type name: string
        """
        self._delete_int_key('backend')  # Set deprecated key to unused state.
        self.set_string('backend-name', name)

    def get_instance_url(self, backend):
        # Dialect 1.2.0 and below used separate keys for each
        # backend-specific setting.
        instance_url = self.get_string(f'{backend}-instance')
        if instance_url:
            return instance_url

        settings = self.backend_settings.get(backend)

        if settings is not None and settings.get('instance-url'):
            return settings.get('instance-url')

        return TRANSLATORS[backend].instance_url

    def set_instance_url(self, backend, instance_url):
        self._delete_str_key(f'{backend}-instance')  # Set deprecated key to unused state.
        self._set_backend_setting(backend, 'instance-url', instance_url)

    def reset_instance_url(self, backend):
        self._delete_str_key(f'{backend}-instance')  # Set deprecated key to unused state.
        self._set_backend_setting(backend, 'instance-url', TRANSLATORS[backend].instance_url)

    def get_dest_langs(self, backend):
        # Dialect 1.2.0 and below used separate keys for each
        # backend-specific setting.
        dest_langs = list(self.get_value(f'{backend}-dest-langs'))
        if dest_langs:
            return dest_langs

        settings = self.backend_settings.get(backend)

        if settings is not None and settings.get('dest-langs'):
            return settings.get('dest-langs')

        return TRANSLATORS[backend].dest_langs

    def set_dest_langs(self, backend, langs):
        self._delete_arr_key(f'{backend}-dest-langs')  # Set deprecated key to unused state.
        self._set_backend_setting(backend, 'dest-langs', langs)

    def reset_dest_langs(self, backend):
        self._delete_arr_key(f'{backend}-dest-langs')  # Set deprecated key to unused state.
        self._set_backend_setting(backend, 'dest-langs', TRANSLATORS[backend].dest_langs)

    def get_src_langs(self, backend):
        # Dialect 1.2.0 and below used separate keys for each
        # backend-specific setting.
        src_langs = list(self.get_value(f'{backend}-src-langs'))
        if src_langs:
            return src_langs

        settings = self.backend_settings.get(backend)

        if settings is not None and settings.get('src-langs'):
            return settings.get('src-langs')

        return TRANSLATORS[backend].src_langs

    def set_src_langs(self, backend, langs):
        self._delete_arr_key(f'{backend}-src-langs')  # Set deprecated key to unused state.
        self._set_backend_setting(backend, 'src-langs', langs)

    def reset_src_langs(self, backend):
        self._delete_arr_key(f'{backend}-src-langs')  # Set deprecated key to unused state.
        self._set_backend_setting(backend, 'src-langs', TRANSLATORS[backend].src_langs)

    @property
    def backend_settings(self):
        return json.loads(self.get_string('backend-settings'))

    @backend_settings.setter
    def backend_settings(self, value):
        self.set_string('backend-settings', json.dumps(value))

    def _set_backend_setting(self, backend, key, value):
        """
        Sets the backend settings with key and value.
        """
        settings = self.backend_settings

        if backend not in settings:
            settings[backend] = {}

        settings[backend][key] = value
        self.backend_settings = settings

    def _delete_arr_key(self, key):
        val = self.get_strv(key)
        if val != []:
            self.set_value(key, GLib.Variant('as', []))

    def _delete_enum_key(self, key):
        val = self.get_enum(key)
        if val > -1:
            self.set_enum(key, -1)

    def _delete_int_key(self, key):
        val = self.get_int(key)
        if val > -1:
            self.set_int(key, -1)

    def _delete_str_key(self, key):
        val = self.get_string(key)
        if val != '':
            self.set_value(
                key,
                GLib.Variant('s', '')
            )



# Copyright 2020 gi-lom
# Copyright 2020-2021 Mufeed Ali
# Copyright 2020-2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later


APP_ID = 'com.github.gi_lom.dialect'
RES_PATH = '/com/github/gi_lom/dialect'
MAX_LENGTH = 5000  # maximum number of characters you can translate at once
TRANS_NUMBER = 10  # number of translations to save in history



# Copyright 2021 Mufeed Ali
# Copyright 2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import httpx

from dialect.translators.basetrans import Detected, TranslatorBase, TranslationError, Translation


class Translator(TranslatorBase):
    name = 'libretranslate'
    prettyname = 'LibreTranslate'
    client = None
    history = []
    languages = []
    supported_features = {
        'mistakes': False,
        'pronunciation': False,
        'change-instance': True,
    }
    instance_url = 'translate.astian.org'

    def __init__(self, base_url=None, **kwargs):
        if base_url is not None:
            self.instance_url = base_url

        self.client = httpx.Client()

        r = self.client.get(self.lang_url)

        for lang in r.json():
            self.languages.append(lang['code'])

    @property
    def detect_url(self):
        if self.instance_url.startswith('localhost:'):
            return 'http://' + self.instance_url + '/detect'
        return 'https://' + self.instance_url + '/detect'

    @property
    def translate_url(self):
        if self.instance_url.startswith('localhost:'):
            return 'http://' + self.instance_url + '/translate'
        return 'https://' + self.instance_url + '/translate'

    @property
    def lang_url(self):
        if self.instance_url.startswith('localhost:'):
            return 'http://' + self.instance_url + '/languages'
        return 'https://' + self.instance_url + '/languages'

    @staticmethod
    def validate_instance_url(url):
        if url.startswith('localhost:'):
            url = 'http://' + url + '/spec'
        else:
            url = 'https://' + url + '/spec'
        client = httpx.Client()
        try:
            r = client.get(url)
            data = r.json()

            if data['info']['title'] == 'LibreTranslate':
                return True

            return False
        except Exception:
            return False

    def detect(self, src_text):
        """Detect the language using the same mechanisms that LibreTranslate uses but locally."""
        try:
            r = self.client.post(
                self.detect_url,
                data={
                    'q': src_text,
                },
            )
            return Detected(r.json()[0]['language'], r.json()[0]['confidence'])
        except Exception as exc:
            raise TranslationError(exc) from exc

    def translate(self, src_text, src, dest):
        try:
            r = self.client.post(
                self.translate_url,
                data={
                    'q': src_text,
                    'source': src,
                    'target': dest,
                },
            )
            return Translation(
                r.json()['translatedText'],
                {
                    'possible-mistakes': None,
                    'src-pronunciation': None,
                    'dest-pronunciation': None,
                },
            )
        except Exception as exc:
            raise TranslationError(exc) from exc



# Copyright 2020 gi-lom
# Copyright 2020-2021 Mufeed Ali
# Copyright 2020-2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import threading
from gettext import gettext as _
from tempfile import NamedTemporaryFile

from gi.repository import Gdk, GLib, GObject, Gst, Gtk, Handy

from dialect.define import APP_ID, MAX_LENGTH, RES_PATH, TRANS_NUMBER
from dialect.lang_selector import DialectLangSelector
from dialect.settings import Settings
from dialect.translators import TRANSLATORS, get_lang_name
from dialect.tts import TTS


@Gtk.Template(resource_path=f'{RES_PATH}/window.ui')
class DialectWindow(Handy.ApplicationWindow):
    __gtype_name__ = 'DialectWindow'

    # Get widgets
    main_stack = Gtk.Template.Child()
    error_message = Gtk.Template.Child()
    translator_box = Gtk.Template.Child()
    retry_backend_btn = Gtk.Template.Child()

    title_stack = Gtk.Template.Child()
    langs_button_box = Gtk.Template.Child()
    switch_btn = Gtk.Template.Child()
    src_lang_btn = Gtk.Template.Child()
    src_lang_label = Gtk.Template.Child()
    dest_lang_btn = Gtk.Template.Child()
    dest_lang_label = Gtk.Template.Child()

    return_btn = Gtk.Template.Child()
    forward_btn = Gtk.Template.Child()

    menu_btn = Gtk.Template.Child()

    src_pron_revealer = Gtk.Template.Child()
    src_pron_label = Gtk.Template.Child()
    mistakes = Gtk.Template.Child()
    mistakes_label = Gtk.Template.Child()
    char_counter = Gtk.Template.Child()
    src_text = Gtk.Template.Child()
    clear_btn = Gtk.Template.Child()
    paste_btn = Gtk.Template.Child()
    src_voice_btn = Gtk.Template.Child()
    translate_btn = Gtk.Template.Child()

    dest_box = Gtk.Template.Child()
    dest_pron_revealer = Gtk.Template.Child()
    dest_pron_label = Gtk.Template.Child()
    dest_text = Gtk.Template.Child()
    trans_spinner = Gtk.Template.Child()
    trans_warning = Gtk.Template.Child()
    copy_btn = Gtk.Template.Child()
    dest_voice_btn = Gtk.Template.Child()

    actionbar = Gtk.Template.Child()
    src_lang_btn2 = Gtk.Template.Child()
    switch_btn2 = Gtk.Template.Child()
    dest_lang_btn2 = Gtk.Template.Child()

    notification_revealer = Gtk.Template.Child()
    notification_label = Gtk.Template.Child()

    translator = None  # Translator object

    # Text to speech
    tts = None
    tts_langs = None
    voice_loading = False  # tts loading status

    # Preset language values
    src_langs = []
    dest_langs = []

    current_history = 0  # for history management

    # Translation-related variables
    no_retranslate = False  # used to prevent unnecessary re-translations
    trans_queue = []  # for pending translations
    active_thread = None  # for ongoing translation
    trans_failed = False  # for monitoring connectivity issues
    trans_mistakes = None  # "mistakes" suggestions
    # Pronunciations
    trans_src_pron = None
    trans_dest_pron = None

    mobile_mode = False  # UI mode

    # Propeties
    backend_loading = GObject.Property(type=bool, default=False)

    def __init__(self, text, langs, **kwargs):
        super().__init__(**kwargs)

        # Options passed to command line
        self.launch_text = text
        self.launch_langs = langs

        # Application object
        self.app = kwargs['application']

        # GStreamer playbin object and related setup
        self.player = Gst.ElementFactory.make('playbin', 'player')
        bus = self.player.get_bus()
        bus.add_signal_watch()
        bus.connect('message', self.on_gst_message)
        self.player_event = threading.Event()  # An event for letting us know when Gst is done playing

        # Clipboard
        self.clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)  # This is only for the Clipboard button

        # Setup window
        self.setup()

    def setup(self):
        self.set_default_icon_name(APP_ID)

        # Load saved dark mode
        gtk_settings = Gtk.Settings.get_default()
        gtk_settings.set_property('gtk-application-prefer-dark-theme',
                                  Settings.get().dark_mode)

        # Connect responsive design function
        self.connect('check-resize', self.responsive_listener)
        # Save settings on close
        self.connect('delete-event', self.save_settings)

        self.setup_headerbar()
        self.setup_actionbar()
        self.setup_translation()
        self.toggle_mobile_mode()

        # Load translator
        self.retry_backend_btn.connect('clicked', self.retry_load_translator)
        threading.Thread(
            target=self.load_translator,
            args=[Settings.get().backend, True],
            daemon=True
        ).start()
        # Get languages available for speech
        if Settings.get().tts != '':
            threading.Thread(target=self.load_lang_speech, daemon=True).start()

    def load_translator(self, backend, launch=False):
        def update_ui():
            # Supported features
            if not self.translator.supported_features['mistakes']:
                self.mistakes.set_revealed(False)

            if not self.translator.supported_features['pronunciation']:
                self.src_pron_revealer.set_reveal_child(False)
                self.dest_pron_revealer.set_reveal_child(False)
                self.app.pronunciation_action.set_enabled(False)
            else:
                self.app.pronunciation_action.set_enabled(True)

            self.no_retranslate = True
            # Update langs list
            self.src_lang_selector.set_languages(self.translator.languages)
            self.dest_lang_selector.set_languages(self.translator.languages)
            # Update selected langs
            self.src_lang_selector.set_property(
                'selected',
                'auto' if Settings.get().src_auto else self._fix_cases(self.src_langs[0])
            )
            self.dest_lang_selector.set_property(
                'selected',
                self._fix_cases(self.dest_langs[0])
            )

            self.no_retranslate = False

            self.main_stack.set_visible_child_name('translate')
            self.set_property('backend-loading', False)

        # Show loading view
        GLib.idle_add(self.main_stack.set_visible_child_name, 'loading')

        try:
            # Translator object
            if TRANSLATORS[backend].supported_features['change-instance']:
                self.translator = TRANSLATORS[backend](
                    base_url=Settings.get().get_instance_url(TRANSLATORS[backend].name)
                )
            else:
                self.translator = TRANSLATORS[backend]()

            # Get saved languages
            self.src_langs = Settings.get().get_src_langs(self.translator.name)
            self.dest_langs = Settings.get().get_dest_langs(self.translator.name)

            # Update UI
            GLib.idle_add(update_ui)

            if launch:
                self.no_retranslate = True
                if self.launch_langs['src'] is not None:
                    self.src_lang_selector.set_property('selected', self.launch_langs['src'])
                if self.launch_langs['dest'] is not None and self.launch_langs['dest'] in self.translator.languages:
                    self.dest_lang_selector.set_property('selected', self.launch_langs['dest'])
                self.no_retranslate = False

                if self.launch_text != '':
                    GLib.idle_add(self.translate, self.launch_text, self.launch_langs['src'], self.launch_langs['dest'])

        except Exception as exc:
            # Show error view
            GLib.idle_add(self.main_stack.set_visible_child_name, 'error')
            GLib.idle_add(self.set_property, 'backend-loading', False)

            self.error_message.set_label(str(exc))
            print('Error: ' + str(exc))

    def retry_load_translator(self, _button):
        threading.Thread(
            target=self.load_translator,
            args=[Settings.get().backend],
            daemon=True
        ).start()

    def on_listen_failed(self):
        self.src_voice_btn.set_image(self.src_voice_warning)
        self.src_voice_spinner.stop()

        self.dest_voice_btn.set_image(self.dest_voice_warning)
        self.dest_voice_spinner.stop()

        tooltip_text = _('A network issue has occured. Retry?')
        self.src_voice_btn.set_tooltip_text(tooltip_text)
        self.dest_voice_btn.set_tooltip_text(tooltip_text)

        self.send_notification(_('A network issue has occured.\nPlease try again.'))

        src_text = self.src_buffer.get_text(
            self.src_buffer.get_start_iter(),
            self.src_buffer.get_end_iter(),
            True
        )
        dest_text = self.dest_buffer.get_text(
            self.dest_buffer.get_start_iter(),
            self.dest_buffer.get_end_iter(),
            True
        )

        if self.tts_langs:
            self.app.listen_src_action.set_enabled(
                self.src_lang_selector.get_property('selected') in self.tts_langs
                and src_text != ''
            )
            self.app.listen_dest_action.set_enabled(
                self.dest_lang_selector.get_property('selected') in self.tts_langs
                and dest_text != ''
            )
        else:
            self.app.listen_src_action.set_enabled(src_text != '')
            self.app.listen_dest_action.set_enabled(dest_text != '')

    def load_lang_speech(self, listen=False, text=None, language=None):
        """
        Load the language list for TTS.

        text and language parameters are only needed with listen parameter.
        """
        try:
            self.voice_loading = True
            self.tts = TTS[Settings.get().tts]()
            self.tts_langs = self.tts.languages
            if not listen:
                GLib.idle_add(self.toggle_voice_spinner, False)
            elif language in self.tts_langs and text != '':
                self.voice_download(text, language)

        except RuntimeError as exc:
            GLib.idle_add(self.on_listen_failed)
            print('Error: ' + str(exc))
        finally:
            if not listen:
                self.voice_loading = False

    def setup_headerbar(self):
        # Left lang selector
        self.src_lang_selector = DialectLangSelector()
        self.src_lang_selector.connect('notify::selected',
                                       self.on_src_lang_changed)
        # Set popover selector to button
        self.src_lang_btn.set_popover(self.src_lang_selector)
        self.src_lang_selector.set_relative_to(self.src_lang_btn)

        # Right lang selector
        self.dest_lang_selector = DialectLangSelector()
        self.dest_lang_selector.connect('notify::selected',
                                        self.on_dest_lang_changed)
        # Set popover selector to button
        self.dest_lang_btn.set_popover(self.dest_lang_selector)
        self.dest_lang_selector.set_relative_to(self.dest_lang_btn)

        self.langs_button_box.set_homogeneous(False)

        # Add menu to menu button
        builder = Gtk.Builder.new_from_resource(f'{RES_PATH}/menu.ui')
        menu = builder.get_object('app-menu')
        menu_popover = Gtk.Popover.new_from_model(self.menu_btn, menu)
        self.menu_btn.set_popover(menu_popover)

    def setup_actionbar(self):
        # Set popovers to lang buttons
        self.src_lang_btn2.set_popover(self.src_lang_selector)
        self.dest_lang_btn2.set_popover(self.dest_lang_selector)

    def setup_translation(self):
        # Left buffer
        self.src_buffer = self.src_text.get_buffer()
        self.src_buffer.set_text(self.launch_text)
        self.src_buffer.connect('changed', self.on_src_text_changed)
        self.src_buffer.connect('end-user-action', self.user_action_ended)
        self.connect('key-press-event', self.update_trans_button)
        # Translate button
        self.translate_btn.connect('clicked', self.translation)
        # "Did you mean" links
        self.mistakes_label.connect('activate-link', self.on_mistakes_clicked)

        # Right buffer
        self.dest_buffer = self.dest_text.get_buffer()
        self.dest_buffer.set_text('')
        self.dest_buffer.connect('changed', self.on_dest_text_changed)
        # Translation progress spinner
        self.trans_spinner.hide()
        self.trans_warning.hide()

        # Voice buttons prep-work
        self.src_voice_warning = Gtk.Image.new_from_icon_name(
            'dialog-warning-symbolic', Gtk.IconSize.BUTTON)
        self.src_voice_image = Gtk.Image.new_from_icon_name(
            'audio-speakers-symbolic', Gtk.IconSize.BUTTON)
        self.src_voice_spinner = Gtk.Spinner()  # For use while audio is running or still loading.

        self.dest_voice_warning = Gtk.Image.new_from_icon_name(
            'dialog-warning-symbolic', Gtk.IconSize.BUTTON)
        self.dest_voice_image = Gtk.Image.new_from_icon_name(
            'audio-speakers-symbolic', Gtk.IconSize.BUTTON)
        self.dest_voice_spinner = Gtk.Spinner()

        self.toggle_voice_spinner(True)

        self.src_voice_btn.set_visible(Settings.get().tts != '')
        self.dest_voice_btn.set_visible(Settings.get().tts != '')

    def responsive_listener(self, _window):
        size = self.get_size()

        if size.width < 680:
            if self.mobile_mode is False:
                self.mobile_mode = True
                self.toggle_mobile_mode()
        else:
            if self.mobile_mode is True:
                self.mobile_mode = False
                self.toggle_mobile_mode()

    def toggle_mobile_mode(self):
        if self.mobile_mode:
            # Show actionbar
            self.actionbar.set_reveal_child(True)
            # Change headerbar title
            self.title_stack.set_visible_child_name('label')
            # Change translation box orientation
            self.translator_box.set_orientation(Gtk.Orientation.VERTICAL)
            # Change lang selectors position
            self.src_lang_selector.set_relative_to(self.src_lang_btn2)
            self.dest_lang_selector.set_relative_to(self.dest_lang_btn2)
        else:
            # Hide actionbar
            self.actionbar.set_reveal_child(False)
            # Reset headerbar title
            self.title_stack.set_visible_child_name('selector')
            # Reset translation box orientation
            self.translator_box.set_orientation(Gtk.Orientation.HORIZONTAL)
            # Reset lang selectors position
            self.src_lang_selector.set_relative_to(self.src_lang_btn)
            self.dest_lang_selector.set_relative_to(self.dest_lang_btn)

    def translate(self, text, src_lang, dest_lang):
        """
        Translates the given text from auto detected language to last used
        language
        """
        # Set src lang to Auto
        if src_lang is None:
            self.src_lang_selector.set_property('selected', 'auto')
        else:
            self.src_lang_selector.set_property('selected', src_lang)
        if dest_lang is not None and dest_lang in self.translator.languages:
            self.dest_lang_selector.set_property('selected', dest_lang)
        # Set text to src buffer
        self.src_buffer.set_text(text)
        # Run translation
        self.translation(None)

    def save_settings(self, *args, **kwargs):
        if not self.is_maximized():
            size = self.get_size()
            Settings.get().window_size = (size.width, size.height)
        if self.translator is not None:
            Settings.get().set_src_langs(self.translator.name, self.src_langs)
            Settings.get().set_dest_langs(self.translator.name, self.dest_langs)

    def send_notification(self, text, timeout=5):
        """
        Display an in-app notification.

        Args:
            text (str): The text or message of the notification.
            timeout (int, optional): The time before the notification disappears. Defaults to 5.
        """
        self.notification_label.set_text(text)
        self.notification_revealer.set_reveal_child(True)

        timer = threading.Timer(
            timeout,
            GLib.idle_add,
            args=[self.notification_revealer.set_reveal_child, False]
        )
        timer.start()

    def toggle_voice_spinner(self, active=True):
        if active:
            self.app.listen_src_action.set_enabled(False)
            self.src_voice_btn.set_image(self.src_voice_spinner)
            self.src_voice_spinner.start()

            self.app.listen_dest_action.set_enabled(False)
            self.dest_voice_btn.set_image(self.dest_voice_spinner)
            self.dest_voice_spinner.start()
        else:
            src_text = self.src_buffer.get_text(
                self.src_buffer.get_start_iter(),
                self.src_buffer.get_end_iter(),
                True
            )
            self.app.listen_src_action.set_enabled(
                self.src_lang_selector.get_property('selected') in self.tts_langs
                and src_text != ''
            )
            self.src_voice_btn.set_image(self.src_voice_image)
            self.src_voice_spinner.stop()

            dest_text = self.dest_buffer.get_text(
                self.dest_buffer.get_start_iter(),
                self.dest_buffer.get_end_iter(),
                True
            )
            self.app.listen_dest_action.set_enabled(
                self.dest_lang_selector.get_property('selected') in self.tts_langs
                and dest_text != ''
            )
            self.dest_voice_btn.set_image(self.dest_voice_image)
            self.dest_voice_spinner.stop()

    def on_src_lang_changed(self, _obj, _param):
        code = self._fix_cases(self.src_lang_selector.get_property('selected'))
        dest_code = self._fix_cases(self.dest_lang_selector.get_property('selected'))
        src_text = self.src_buffer.get_text(
            self.src_buffer.get_start_iter(),
            self.src_buffer.get_end_iter(),
            True
        )

        if code == dest_code:
            code = self._fix_cases(
                self.dest_langs[1] if code == self.src_langs[0] else dest_code
            )
            self.dest_lang_selector.set_property('selected', self._fix_cases(self.src_langs[0]))

        # Disable or enable listen function.
        if self.tts_langs and Settings.get().tts != '':
            self.app.listen_src_action.set_enabled(code in self.tts_langs
                                                   and src_text != '')

        if code in self.translator.languages:
            self.src_lang_label.set_label(get_lang_name(code))
            # Update saved src langs list
            if code in self.src_langs:
                # Bring lang to the top
                self.src_langs.remove(code)
            elif code.lower() in self.src_langs:
                # Bring lang to the top
                self.src_langs.remove(code.lower())
            elif len(self.src_langs) == 4:
                self.src_langs.pop()
            self.src_langs.insert(0, code)
        else:
            self.src_lang_label.set_label(_('Auto'))

        # Rewrite recent langs
        self.src_lang_selector.clear_recent()
        self.src_lang_selector.insert_recent('auto', _('Auto'))
        for code in self.src_langs:
            code = self._fix_cases(code)
            self.src_lang_selector.insert_recent(code, get_lang_name(code))

        # Refresh list
        self.src_lang_selector.refresh_selected()

        # Translate again
        if not self.no_retranslate:
            self.translation(None)

    def on_dest_lang_changed(self, _obj, _param):
        code = self._fix_cases(self.dest_lang_selector.get_property('selected'))
        src_code = self._fix_cases(self.src_lang_selector.get_property('selected'))
        dest_text = self.dest_buffer.get_text(
            self.dest_buffer.get_start_iter(),
            self.dest_buffer.get_end_iter(),
            True
        )

        if code == src_code:
            code = self._fix_cases(src_code)
            self.src_lang_selector.set_property('selected', self._fix_cases(self.dest_langs[0]))

        # Disable or enable listen function.
        if self.tts_langs and Settings.get().tts != '':
            self.app.listen_dest_action.set_enabled(code in self.tts_langs
                                                    and dest_text != '')

        self.dest_lang_label.set_label(get_lang_name(code))
        # Update saved dest langs list
        if code in self.dest_langs:
            # Bring lang to the top
            self.dest_langs.remove(code)
        elif code.lower() in self.dest_langs:
            # Bring lang to the top
            self.dest_langs.remove(code.lower())
        elif len(self.src_langs) == 4:
            self.dest_langs.pop()
        self.dest_langs.insert(0, code)

        # Rewrite recent langs
        self.dest_lang_selector.clear_recent()
        for code in self.dest_langs:
            code = self._fix_cases(code)
            self.dest_lang_selector.insert_recent(code, get_lang_name(code))

        # Refresh list
        self.dest_lang_selector.refresh_selected()

        # Translate again
        if not self.no_retranslate:
            self.translation(None)

    """
    User interface functions
    """
    def ui_return(self, _action, _param):
        """Go back one step in history."""
        if self.current_history != TRANS_NUMBER:
            self.current_history += 1
            self.history_update()

    def ui_forward(self, _action, _param):
        """Go forward one step in history."""
        if self.current_history != 0:
            self.current_history -= 1
            self.history_update()

    def add_history_entry(self, src_language, dest_language, src_text, dest_text):
        """Add a history entry to the history list."""
        new_history_trans = {
            'Languages': [src_language, dest_language],
            'Text': [src_text, dest_text]
        }
        if self.current_history > 0:
            del self.translator.history[: self.current_history]
            self.current_history = 0
        if len(self.translator.history) == TRANS_NUMBER:
            self.translator.history.pop()
        self.translator.history.insert(0, new_history_trans)
        GLib.idle_add(self.reset_return_forward_btns)

    def switch_all(self, src_language, dest_language, src_text, dest_text):
        self.src_lang_selector.set_property('selected', dest_language)
        self.dest_lang_selector.set_property('selected', src_language)
        self.src_buffer.set_text(dest_text)
        self.dest_buffer.set_text(src_text)
        self.add_history_entry(src_language, dest_language, src_text, dest_text)

        # Re-enable widgets
        self.langs_button_box.set_sensitive(True)
        self.translate_btn.set_sensitive(self.src_buffer.get_char_count() != 0)

    def switch_auto_lang(self, dest_language, src_text, dest_text):
        src_language = self.translator.detect(src_text).lang
        if isinstance(src_language, list):
            src_language = src_language[0]

        # Switch all
        GLib.idle_add(self.switch_all, src_language, dest_language, src_text, dest_text)

    def ui_switch(self, _action, _param):
        # Get variables
        self.langs_button_box.set_sensitive(False)
        self.translate_btn.set_sensitive(False)
        src_language = self.src_lang_selector.get_property('selected')
        dest_language = self.dest_lang_selector.get_property('selected')
        src_text = self.src_buffer.get_text(
            self.src_buffer.get_start_iter(),
            self.src_buffer.get_end_iter(),
            True
        )
        dest_text = self.dest_buffer.get_text(
            self.dest_buffer.get_start_iter(),
            self.dest_buffer.get_end_iter(),
            True
        )
        if src_language == 'auto':
            if src_text == '':
                src_language = self.src_langs[0]
            else:
                threading.Thread(
                    target=self.switch_auto_lang,
                    args=(dest_language, src_text, dest_text),
                    daemon=True
                ).start()
                return

        # Switch all
        self.switch_all(src_language, dest_language, src_text, dest_text)

    def ui_clear(self, _action, _param):
        self.src_buffer.set_text('')
        self.src_buffer.emit('end-user-action')

    def ui_copy(self, _action, _param):
        dest_text = self.dest_buffer.get_text(
            self.dest_buffer.get_start_iter(),
            self.dest_buffer.get_end_iter(),
            True
        )
        self.clipboard.set_text(dest_text, -1)
        self.clipboard.store()

    def ui_paste(self, _action, _param):
        text = self.clipboard.wait_for_text()
        if text is not None:
            end_iter = self.src_buffer.get_end_iter()
            self.src_buffer.insert(end_iter, text)

    def ui_src_voice(self, _action, _param):
        src_text = self.src_buffer.get_text(
            self.src_buffer.get_start_iter(),
            self.src_buffer.get_end_iter(),
            True
        )
        src_language = self.src_lang_selector.get_property('selected')
        self._voice(src_text, src_language)

    def ui_dest_voice(self, _action, _param):
        dest_text = self.dest_buffer.get_text(
            self.dest_buffer.get_start_iter(),
            self.dest_buffer.get_end_iter(),
            True
        )
        dest_language = self.dest_lang_selector.get_property('selected')
        self._voice(dest_text, dest_language)

    def _voice(self, text, lang):
        if text != '':
            self.toggle_voice_spinner(True)
            if self.tts_langs:
                threading.Thread(
                    target=self.voice_download,
                    args=(text, lang),
                    daemon=True
                ).start()
            else:
                threading.Thread(
                    target=self.load_lang_speech,
                    args=(True, text, lang),
                    daemon=True
                ).start()

    def on_gst_message(self, _bus, message):
        if message.type == Gst.MessageType.EOS:
            self.player.set_state(Gst.State.NULL)
            self.player_event.set()
        elif message.type == Gst.MessageType.ERROR:
            self.player.set_state(Gst.State.NULL)
            self.player_event.set()
            print('Some error occured while trying to play.')

    def voice_download(self, text, language):
        try:
            self.voice_loading = True

            with NamedTemporaryFile() as file_to_play:
                self.tts.download_voice(text, language, file_to_play)
                self.player.set_property('uri', 'file://' + file_to_play.name)
                self.player.set_state(Gst.State.PLAYING)
                self.player_event.wait()
        except Exception as exc:
            print(exc)
            print('Audio download failed.')
            GLib.idle_add(self.on_listen_failed)
        else:
            GLib.idle_add(self.toggle_voice_spinner, False)
        finally:
            self.voice_loading = False

    # This starts the translation if Ctrl+Enter button is pressed
    def update_trans_button(self, button, keyboard):
        modifiers = keyboard.get_state() & Gtk.accelerator_get_default_mod_mask()

        control_mask = Gdk.ModifierType.CONTROL_MASK
        shift_mask = Gdk.ModifierType.SHIFT_MASK
        unicode_key_val = Gdk.keyval_to_unicode(keyboard.keyval)
        enter_keys = (Gdk.KEY_Return, Gdk.KEY_KP_Enter)
        if (GLib.unichar_isgraph(chr(unicode_key_val)) and
                modifiers in (shift_mask, 0) and not self.src_text.is_focus()):
            self.src_text.grab_focus()

        if not Settings.get().live_translation:
            if control_mask == modifiers:
                if keyboard.keyval in enter_keys:
                    if not Settings.get().translate_accel_value:
                        self.translation(button)
                        return Gdk.EVENT_STOP
                    return Gdk.EVENT_PROPAGATE
            elif keyboard.keyval in enter_keys:
                if Settings.get().translate_accel_value:
                    self.translation(button)
                    return Gdk.EVENT_STOP
                return Gdk.EVENT_PROPAGATE

        return Gdk.EVENT_PROPAGATE

    def on_mistakes_clicked(self, _button, _data):
        self.mistakes.set_revealed(False)
        self.src_buffer.set_text(self.trans_mistakes[1])
        # Run translation again
        self.translation(None)

    def on_src_text_changed(self, buffer):
        sensitive = buffer.get_char_count() != 0
        self.translate_btn.set_sensitive(sensitive)
        self.app.clear_action.set_enabled(sensitive)
        if not self.voice_loading and self.tts_langs:
            self.app.listen_src_action.set_enabled(
                self.src_lang_selector.get_property('selected') in self.tts_langs
                and sensitive
            )
        elif not self.voice_loading and not self.tts_langs:
            self.app.listen_src_action.set_enabled(sensitive)

    def on_dest_text_changed(self, buffer):
        sensitive = buffer.get_char_count() != 0
        self.app.copy_action.set_enabled(sensitive)
        if not self.voice_loading and self.tts_langs:
            self.app.listen_dest_action.set_enabled(
                self.dest_lang_selector.get_property('selected') in self.tts_langs
                and sensitive
            )
        elif not self.voice_loading and not self.tts_langs:
            self.app.listen_dest_action.set_enabled(sensitive)

    def user_action_ended(self, buffer):
        # If the text is over the highest number of characters allowed, it is truncated.
        # This is done for avoiding exceeding the limit imposed by Google.
        if buffer.get_char_count() >= MAX_LENGTH:
            self.send_notification(_('5000 characters limit reached!'))
            src_text = buffer.get_text(
                buffer.get_start_iter(),
                buffer.get_end_iter(),
                True
            )
            self.src_buffer.set_text(src_text[:MAX_LENGTH])
        self.char_counter.set_text(f'{str(buffer.get_char_count())}/{MAX_LENGTH}')
        if Settings.get().live_translation:
            self.translation(None)

    # The history part
    def reset_return_forward_btns(self):
        self.app.back_action.set_enabled(self.current_history < len(self.translator.history) - 1)
        self.app.forward_action.set_enabled(self.current_history > 0)

    # Retrieve translation history
    def history_update(self):
        self.reset_return_forward_btns()
        lang_hist = self.translator.history[self.current_history]
        self.no_retranslate = True
        self.src_lang_selector.set_property('selected',
                                            lang_hist['Languages'][0])
        self.dest_lang_selector.set_property('selected',
                                             lang_hist['Languages'][1])
        self.no_retranslate = False
        self.src_buffer.set_text(lang_hist['Text'][0])
        self.dest_buffer.set_text(lang_hist['Text'][1])

    def set_no_retranslate(self, state):
        self.no_retranslate = state

    # THE TRANSLATION AND SAVING TO HISTORY PART
    def appeared_before(self):
        src_language = self.src_lang_selector.get_property('selected')
        dest_language = self.dest_lang_selector.get_property('selected')
        src_text = self.src_buffer.get_text(
            self.src_buffer.get_start_iter(),
            self.src_buffer.get_end_iter(),
            True
        )
        if (
            len(self.translator.history) >= self.current_history + 1
            and (self.translator.history[self.current_history]['Languages'][0] == src_language or 'auto')
            and self.translator.history[self.current_history]['Languages'][1] == dest_language
            and self.translator.history[self.current_history]['Text'][0] == src_text
            and not self.trans_failed
        ) or (
            len(self.trans_queue) == 1
            and (self.trans_queue[0].get('src_language') == src_language or 'auto')
            and self.trans_queue[0].get('dest_language') == dest_language
            and self.trans_queue[0].get('src_text') == src_text
        ):
            return True
        return False

    def translation(self, _button):
        # If it's like the last translation then it's useless to continue
        if not self.appeared_before():
            src_text = self.src_buffer.get_text(
                self.src_buffer.get_start_iter(),
                self.src_buffer.get_end_iter(),
                True
            )
            src_language = self.src_lang_selector.get_property('selected')
            dest_language = self.dest_lang_selector.get_property('selected')

            if self.trans_queue:
                self.trans_queue.pop(0)
            self.trans_queue.append({
                'src_text': src_text,
                'src_language': src_language,
                'dest_language': dest_language
            })

            # Check if there are any active threads.
            if self.active_thread is None:
                # Show feedback for start of translation.
                self.trans_spinner.show()
                self.trans_spinner.start()
                self.dest_box.set_sensitive(False)
                self.langs_button_box.set_sensitive(False)
                # If there is no active thread, create one and start it.
                self.active_thread = threading.Thread(target=self.run_translation, daemon=True)
                self.active_thread.start()

    def change_backends(self, backend):
        self.set_property('backend-loading', True)

        # Save previous backend settings
        self.save_settings()

        # Load translator
        threading.Thread(
            target=self.load_translator,
            args=[backend],
            daemon=True
        ).start()

    def run_translation(self):
        def on_trans_failed():
            self.trans_warning.show()
            self.send_notification(_('Translation failed.\nPlease check for network issues.'))
            self.app.copy_action.set_enabled(False)
            self.app.listen_src_action.set_enabled(False)
            self.app.listen_dest_action.set_enabled(False)

        def on_trans_success():
            self.trans_warning.hide()

        def on_trans_done():
            self.trans_spinner.stop()
            self.trans_spinner.hide()
            self.dest_box.set_sensitive(True)
            self.langs_button_box.set_sensitive(True)

        def on_mistakes():
            if self.trans_mistakes is not None and self.translator.supported_features['mistakes']:
                self.mistakes_label.set_markup(_('Did you mean: ') + f'<a href="#">{self.trans_mistakes[0]}</a>')
                self.mistakes.set_revealed(True)
            elif self.mistakes.get_revealed():
                self.mistakes.set_revealed(False)

        def on_pronunciation():
            reveal = Settings.get().show_pronunciation
            if self.translator.supported_features['pronunciation']:
                if self.trans_src_pron is not None:
                    self.src_pron_label.set_text(self.trans_src_pron)
                    self.src_pron_revealer.set_reveal_child(reveal)
                elif self.src_pron_revealer.get_reveal_child():
                    self.src_pron_revealer.set_reveal_child(False)

                if self.trans_dest_pron is not None:
                    self.dest_pron_label.set_text(self.trans_dest_pron)
                    self.dest_pron_revealer.set_reveal_child(reveal)
                elif self.dest_pron_revealer.get_reveal_child():
                    self.dest_pron_revealer.set_reveal_child(False)

        while self.trans_queue:
            # If the first language is revealed automatically, let's set it
            trans_dict = self.trans_queue.pop(0)
            src_text = trans_dict['src_text']
            src_language = trans_dict['src_language']
            dest_language = trans_dict['dest_language']
            if src_language == 'auto' and src_text != '':
                try:
                    src_language = self.translator.detect(src_text).lang
                    if isinstance(src_language, list):
                        src_language = src_language[0]
                    if src_language in self.translator.languages:
                        GLib.idle_add(self.set_no_retranslate, True)
                        GLib.idle_add(self.src_lang_selector.set_property,
                                      'selected', src_language)
                        GLib.idle_add(self.set_no_retranslate, False)
                        if src_language not in self.src_langs:
                            self.src_langs[0] = src_language
                    else:
                        src_language = 'auto'
                except Exception:
                    self.trans_failed = True
            # If the two languages are the same, nothing is done
            if src_language != dest_language:
                dest_text = ''
                # THIS IS WHERE THE TRANSLATION HAPPENS. The try is necessary to circumvent a bug of the used API
                if src_text != '':
                    try:
                        translation = self.translator.translate(
                            src_text,
                            src=src_language,
                            dest=dest_language
                        )
                        dest_text = translation.text
                        self.trans_mistakes = translation.extra_data['possible-mistakes']
                        self.trans_src_pron = translation.extra_data['src-pronunciation']
                        self.trans_dest_pron = translation.extra_data['dest-pronunciation']
                        self.trans_failed = False
                    except Exception as exc:
                        print(exc)
                        self.trans_mistakes = None
                        self.trans_pronunciation = None
                        self.trans_failed = True

                    # Finally, everything is saved in history
                    self.add_history_entry(
                        src_language,
                        dest_language,
                        src_text,
                        dest_text
                    )
                else:
                    self.trans_failed = False
                    self.trans_mistakes = None
                    self.trans_src_pron = None
                    self.trans_dest_pron = None
                GLib.idle_add(self.dest_buffer.set_text, dest_text)
                GLib.idle_add(on_mistakes)
                GLib.idle_add(on_pronunciation)

        if self.trans_failed:
            GLib.idle_add(on_trans_failed)
        else:
            GLib.idle_add(on_trans_success)
        GLib.idle_add(on_trans_done)
        self.active_thread = None

    @staticmethod
    def _fix_cases(code):
        return code.replace('cn', 'CN').replace('tw', 'TW')
